#ifdef DEBUG
INSERT_VARIABLES_HERE

uniform vec2 debugCamRotation;
uniform vec3 debugCamPosition;

VAR bool debugPlaneShown = false;
VAR float debugPlaneHeight = 0.0;

VAR bool raymarchShowExpired = false;
VAR int raymarchDebugColors = 0;
VAR int raymarchMaxIterations = 99;
VAR float raymarchMaxDepth = 50.0;
#else
#define raymarchMaxIterations 120
#define raymarchMaxDepth 200.0
#endif

uniform float t;

const float PI = 3.14159265;
//const float TAU = 2. * PI;

const int RAY_EXPIRED = 0;
const int RAY_HIT = 1;
const int RAY_ESCAPED = 2;

struct RaymarchResult {
    float depth;
    vec4 materialParam;
};
RaymarchResult makeRR(float depth, float materialId) {
    return RaymarchResult(depth, vec4(materialId));
}

RaymarchResult makeRR(float depth, float materialId, vec3 param) {
    return RaymarchResult(depth, vec4(param, materialId));
}

struct HitInfo {
    float depth;
    vec4 materialParam;
    int rayHit;
#ifdef DEBUG
    int numIterations;
#endif
    vec3 finalPos;
};
// float sceneSwitch(); // inlined for size savings
vec3 camFunc(inout vec3 p);
vec3 calcMaterialColor(vec4 material, out vec2 materialSpecularity);
RaymarchResult sceneSDF(vec3 p);

RaymarchResult objUnion(RaymarchResult old, RaymarchResult new) {
    if (new.depth < old.depth) {
        return new;
    }
    return old;
}

#ifdef DEBUG
float linePlaneIntersection(vec3 linePoint, vec3 lineDir, vec3 planePoint, vec3 planeNormal) {
    return (dot(planePoint - linePoint, planeNormal)) / dot(lineDir, planeNormal);
}

vec3 fusion(float x) {
    float t = clamp(x,0.0,1.0);
    return clamp(vec3(sqrt(t), t*t*t, max(sin(PI*1.75*t), pow(t, 12.0))), 0.0, 1.0);
}

// HDR version
vec3 fusionHDR(float x) {
    float t = clamp(x,0.0,1.0);
    return fusion(sqrt(t))*(0.5+2.*t);
}

vec3 distanceMeter(float dist, float rayLength, vec3 rayDir, float camDistance) {
    float idealGridDistance = 20.0/rayLength*pow(abs(rayDir.y),0.8);
    float nearestBase = floor(log(idealGridDistance)/log(10.));
    float relativeDist = abs(dist / camDistance);

    float largerDistance = pow(10.0,nearestBase+1.);
    float smallerDistance = pow(10.0,nearestBase);

    vec3 col = fusionHDR(log(1.+relativeDist));
    col = max(vec3(0.),col);
    if (sign(dist) < 0.) {
        col = col.grb * 3.;
    }

    float l0 = (pow(0.5+0.5*cos(dist*PI*2.*smallerDistance),10.0));
    float l1 = (pow(0.5+0.5*cos(dist*PI*2.*largerDistance),10.0));

    float x = fract(log(idealGridDistance) / log(10.));
    l0 = mix(l0,0.,smoothstep(0.5,1.0,x));
    l1 = mix(0.,l1,smoothstep(0.0,0.5,x));

    col.rgb *= 0.1 + 0.9 * (1. - l0) * (1. - l1);
    return col;
}
#endif

// TODO add hash functions from https://www.shadertoy.com/view/4djSRW
//----------------------------------------------------------------------------------------
vec3 hashOld33( vec3 p )
{
	p = vec3( dot(p,vec3(127.1,311.7, 74.7)),
			  dot(p,vec3(269.5,183.3,246.1)),
			  dot(p,vec3(113.5,271.9,124.6)));

	return fract(sin(p)*43758.5453123);
}

float tri(float x) {
    float v2 = fract(x) * 4.;
    if (v2 < 2.) return v2 - 1.;
    return 3. - v2;
}

#if 0
//
// Repeat space along one axis. Use like this to repeat along the x axis:
// <float cell = pMod1(p.x,5);> - using the return value is optional.
float pMod1(inout float p, float size) {
    float halfsize = size*0.5;
    float c = floor((p + halfsize)/size);
    p = mod(p + halfsize, size) - halfsize;
    return c;
}
float vmax(vec2 v) {
    return max(v.x, v.y);
}

float vmax(vec4 v) {
    return max(max(v.x, v.y), max(v.z, v.w));
}

float vmin(vec2 v) {
    return min(v.x, v.y);
}

float vmin(vec3 v) {
    return min(min(v.x, v.y), v.z);
}

float vmin(vec4 v) {
    return min(min(v.x, v.y), min(v.z, v.w));
}

// Cheap Box: distance to corners is overestimated
float fBoxCheap(vec3 p, vec3 b) {
    return vmax(abs(p) - b);
}

// Endless box in two dimensions
float fBox2Cheap(vec2 p, vec2 b) {
        return vmax(abs(p)-b);
}

float fBox2(vec2 p, vec2 b) {
        vec2 d = abs(p) - b;
        return length(max(d, vec2(0))) + vmax(min(d, vec2(0)));
}

// Cylinder standing upright on the xz plane
float fCylinder(vec3 p, float r, float height) {
    float d = length(p.xz) - r;
    d = max(d, abs(p.y) - height);
    return d;
}

// Endless cylinder in two dimensions
float fCylinder2(vec2 p, float r) {
    return length(p) - r;
}

float fTrianglePrism(vec3 p, float size, float depth) {
    vec3 q = abs(p);
    return max(q.z - depth, max(q.x * 0.866025 + p.y * 0.5, -p.y) - size * 0.5);
}

float fTrianglePrismUpright(vec3 p, float size, float depth) {
    vec3 q = abs(p);
    return max(q.y - depth, max(q.x * 0.866025 + p.z * 0.5, -p.z) - size * 0.5);
}

float fPyramid(vec3 p, vec3 h ) {
    // h = { cos a, sin a, height }
    // Tetrahedron = Octahedron - Cube
    float box = fBox(p - vec3(0., -2.0 * h.z, 0), vec3(2. * h.z));

    float d = 0.0;
    d = max(d, abs(dot(p, vec3(-h.x, h.y,  0 ))));
    d = max(d, abs(dot(p, vec3( h.x, h.y,  0 ))));
    d = max(d, abs(dot(p, vec3( 0,   h.y,  h.x ))));
    d = max(d, abs(dot(p, vec3( 0,   h.y, -h.x ))));
    float octa = d - h.z;
    return max(-box, octa);
}

// Torus in the XZ-plane
float fTorus(vec3 p, float smallRadius, float largeRadius) {
    return length(vec2(length(p.xz) - largeRadius, p.y)) - smallRadius;
}

float fSphere(vec3 p, float r) {
    return length(p) - r;
}

float fOpIntersect(float a, float b) {
    return max(a, b);
}

vec2 rotate90(vec2 p) {
    return vec2(p.y, 1. - p.x);
}


// Repeat in three dimensions
vec3 pMod3(inout vec3 p, vec3 size) {
    vec3 c = floor((p + size*0.5)/size);
    p = mod(p + size*0.5, size) - size*0.5;
    return c;
}

// Repeat around the origin by a fixed angle.
// For easier use, num of repetitions is use to specify the angle.
float pModPolar(inout vec2 p, float repetitions) {
    float angle = 2.*PI/repetitions;
    float a = atan(p.y, p.x) + angle/2.;
    float r = length(p);
    float c = floor(a/angle);
    a = mod(a,angle) - angle/2.;
    p = vec2(cos(a), sin(a))*r;
    // For an odd number of repetitions, fix cell index of the cell in -x direction
    // (cell index would be e.g. -5 and 5 in the two halves of the cell):
    if (abs(c) >= (repetitions/2.)) c = abs(c);
    return c;
}


#endif

mat3 lookAt(vec3 eye, vec3 target/*, float rotation*/) {
    vec3 cw = normalize(target - eye);
    /*vec3 cp = vec3(sin(rotation), cos(rotation),0.0);*/
    vec3 cu = normalize( cross(cw,/*cp*/ vec3(0,1,0) ));
    vec3 cv =          ( cross(cu,cw) );
    return mat3( cu, cv, cw );
}

float fOpUnion(float a, float b) {
    return min(a, b);
}

float smin(float a, float b, float k) {
    float h = max(k - abs(a - b), 0.0);
    return min(a, b) - h*h*0.25/k;
}

float vmax(vec3 v) {
    return max(max(v.x, v.y), v.z);
}

// Box: correct distance to corners
float fBox(vec3 p, vec3 b) {
    vec3 d = abs(p) - b;
    return length(max(d, vec3(0))) + vmax(min(d, vec3(0)));
}

// Cylinder standing upright on the xz plane
float fCylinder(vec3 p, float r, float height) {
    float d = length(p.xz) - r;
    d = max(d, abs(p.y) - height);
    return d;
}

// Plane with normal n (n is normalized) at some distance from the origin
float fPlane(vec3 p, vec3 n, float distanceFromOrigin) {
    return dot(p, n) + distanceFromOrigin;
}

// Capsule: A Cylinder with round caps on both sides
float fCapsule(vec3 p, float r, float c) {
        return mix(length(p.xz) - r, length(vec3(p.x, abs(p.y) - c, p.z)) - r, step(c, abs(p.y)));
}

//float fOpDifference(float a, float b) {
//    return max(a,-b);
//}
#define fOpDifference(a,b) max(a, -b)

void pRotate(inout vec2 p, float a) {
    p = cos(a) * p + sin(a) * vec2(p.y, -p.x);
}

// Repeat in two dimensions
vec2 pMod2(inout vec2 p, vec2 size) {
        vec2 c = floor((p + size*0.5)/size);
        p = mod(p + size*0.5,size) - size*0.5;
        return c;
}

vec3 sceneNormal(vec3 p) {
#if 0
    float e = .01 + min(t, 0.0);
    return normalize(vec3(
        sceneSDF(p + vec3(e, 0, 0)).depth - sceneSDF(p - vec3(e, 0, 0)).depth,
        sceneSDF(p + vec3(0, e, 0)).depth - sceneSDF(p - vec3(0, e, 0)).depth,
        sceneSDF(p + vec3(0, 0, e)).depth - sceneSDF(p - vec3(0, 0, e)).depth));
#elif 0
    float e = .01;
    vec3 n = vec3(0.0);
    vec3 v = vec3(1,0,0);
    for(int i = 0; i < 3; i++) {
        n += v * (sceneSDF(p + e * v).depth - sceneSDF(p - e * v).depth);
        v = v.zxy;
    }
    return normalize(n);
#else
    // prevent the compiler from inlining sceneSDF() 4 times
    float ep = 0.1;
    vec3 n = vec3(0.0);
    for(int i=0; i<4; i++)
    {
        vec3 e = 0.5773*(2.0*vec3(i==0||i==3,i>=2,i==1||i==3)-1.0);
        n += e*sceneSDF(p+e*ep).depth;
        if (i == int(7. + min(t, 0.)))
            break;
    }
    return normalize(n);
#endif
}

HitInfo castRay(vec3 eye, vec3 ray) {
    int numIterations = 0;
    float depth = 0.0;
    vec3 finalPos;
    HitInfo hitInfo;
    RaymarchResult result;

    // Stupid GLSL compiler needs a constant limit here...
    for (int i = 0; i < 99999999; i++) {
        if (i > raymarchMaxIterations) {
            hitInfo.rayHit = RAY_EXPIRED;
            break;
        }

        numIterations++;
        finalPos = eye + depth * ray;
        result = sceneSDF(finalPos);

        float dist = result.depth;

        // big brain stuff by iq here:
        //  - abs() -> can help with artifacts if non-exact sdf steps too much and goes inside geometry
        //  - scale depth comparion according to depth -> as details get smaller on screen (due to bigger distance) we care less
        if (abs(dist) < 1e-4 * depth) {
            hitInfo.rayHit = RAY_HIT;
            break;
        }
        depth += dist;

        if (depth >= raymarchMaxDepth) {
            hitInfo.rayHit = RAY_ESCAPED;
            break;
        }
    }
    hitInfo.depth = depth;
    hitInfo.materialParam = result.materialParam;
    hitInfo.finalPos = finalPos;
#ifdef DEBUG
    hitInfo.numIterations = numIterations;
#endif
    return hitInfo;
}

void main() {
    vec2 pos = 2.0 * (gl_FragCoord.xy / vec2(CANVAS_WIDTH, CANVAS_HEIGHT) - 0.5); // normalize to -1..1
    pos.x *= CANVAS_WIDTH / CANVAS_HEIGHT; // aspect ratio tweak
    pos *= radians(60.0) / 2.0; // FOV
    vec3 ray = normalize(vec3(pos, 1.));
#ifdef DEBUG
    vec3 eye = debugCamPosition;
    pRotate(ray.yz, radians(-debugCamRotation.y));
    pRotate(ray.xz, radians(debugCamRotation.x));
#else
    vec3 eye = camFunc(ray);
#endif

    int numIterations = 0;
    HitInfo hitInfo = castRay(eye, ray);
    vec3 color = vec3(0.);

#ifdef DEBUG
    bool hitAxes = false;
    bool hitDebugPlane = false;
    if (true) {
        const float AXIS_WIDTH = 0.03;
        float dXZ = linePlaneIntersection(eye, ray, vec3(0.), vec3(0., 1., 0.));
        vec3 pXZ = eye + dXZ * ray;
        if (dXZ > 0.0 && (hitInfo.rayHit == RAY_ESCAPED || dXZ < hitInfo.depth)) {
            if (abs(pXZ.z) < AXIS_WIDTH) {
                hitAxes = true;
                color = vec3(1., 0., 0.);
            }
            if (abs(pXZ.x) < AXIS_WIDTH) {
                hitAxes = true;
                color = vec3(0., 0., 1.);
            }
        }
        float dXY = linePlaneIntersection(eye, ray, vec3(0.), vec3(0., 0., 1.));
        vec3 pXY = eye + dXY * ray;
        if (dXY > 0.0 && (hitInfo.rayHit == RAY_ESCAPED || dXY < hitInfo.depth)) {
            if (abs(pXY.x) < AXIS_WIDTH) {
                hitAxes = true;
                color = vec3(0., 1., 0.);
            }
        }
    }

    if (debugPlaneShown) {
        // Calculate line-plane intersection
        vec3 debugPlaneNormal = vec3(0., 1., 0.);
        vec3 debugPlanePoint = vec3(0., debugPlaneHeight, 0.);
        float d = linePlaneIntersection(eye, ray, debugPlanePoint, debugPlaneNormal);
        if (d > 0.0 && (hitInfo.rayHit == RAY_ESCAPED || d < hitInfo.depth)) {
            hitDebugPlane = true;
            hitInfo.finalPos = eye + d * ray;
        }
    }
#endif

    vec2 materialSpecularity;
    vec3 material = calcMaterialColor(hitInfo.materialParam, materialSpecularity);
    vec3 normal = sceneNormal(hitInfo.finalPos);

    // Tweak sky color based on screen pos (technique from iq)
    vec3 backgroundColor = vec3(0.282, 0.833, 0.940);
    backgroundColor -= 0.5 * ray.y;

#ifdef DEBUG
    if (hitAxes) {
        // already calculated above
    } else if (hitDebugPlane) {
        // WTF is this calculation?
        float t = pow(2., 8.);
        //if (ray.y < 0.) {
            t = eye.y / -ray.y;
        //}
        // FIXME add tonemap + gamma correct: https://www.shadertoy.com/view/ldK3zD
        // ??? what is t?
        // orig shader calls: distanceMeter(dist, t, d, cam_pos.y)
        // where dist = return value from SDF, t = cam.y/-ray.y, d = ray direction
        float debugDist = sceneSDF(hitInfo.finalPos).depth;
        color = distanceMeter(debugDist, t, ray, length(eye - hitInfo.finalPos));
    } else if (hitInfo.rayHit == RAY_EXPIRED && raymarchShowExpired) {
        color = vec3(1.0, 0.0, 1.0);
    } else if (raymarchDebugColors == 1) {
        // shade based on num iterations
        float amt = float(hitInfo.numIterations) / float(raymarchMaxIterations);
        if (hitInfo.rayHit == RAY_HIT)
            color = vec3(amt, 0.0, 0.0);
        else
            color = vec3(0.0, amt, 0.0);
    } else if (raymarchDebugColors == 2) {
        // shade based on surface normal
        if (hitInfo.rayHit == RAY_HIT)
            color = (normal + 1.0) / 2.0;
    } else
#endif
    {
        if (hitInfo.rayHit == RAY_ESCAPED) {
            color = backgroundColor;
        } else {
#ifdef BLINN_PHONG_LIGHTS
            color = vec3(0.0, 1.0, 1.0);
            const vec3 lightPos = vec3(-3, 10, -2);
            const vec3 lightColor = vec3(1.0, 1.0, 1.0);
            const float lightPower = 1.0;
            const vec3 ambientColor = vec3(0.04, 0.04, 0.04);
            const vec3 diffuseColor = vec3(0.8, 0.8, 0.8);
            const vec3 specColor = vec3(1.0, 1.0, 1.0);
            const float shininess = 64.0;

            vec3 lightDir = normalize(lightPos - hitInfo.finalPos);
            float lightDistSq = dot(lightDir, lightDir);

            float lambertian = max(dot(lightDir, normal), 0.0);

            vec3 halfDir = normalize(lightDir - ray);
            float specAngle = max(dot(halfDir, normal), 0.0);
            float specular = pow(specAngle, shininess);

            vec3 linearColor = ambientColor +
                diffuseColor * lambertian * lightColor * lightPower / lightDistSq +
                specColor * specular * lightColor * lightPower / lightDistSq;
            color = material * linearColor;
            color = pow(color, vec3(1. / 2.2)); // gamma correct
#else
            vec3 sunColor = vec3(0.470, 0.433, 0.202);
            float sunColorMult = 4.456;
            sunColor *= sunColorMult;
            vec3 sunSpecularColor = vec3(0.8, 0.90, 0.60);
            sunSpecularColor *= sunColorMult;
            float sunDirectionX = 0.46;
            float sunDirectionY = 0.7;
            float sunDirectionZ = 0.84;
            //vec3 sunDirection = normalize(vec3(-0.68, 0.47, 0.67));
            vec3 sunDirection = normalize(vec3(sunDirectionX, sunDirectionY, -sunDirectionZ));
            float sunShininess = 8.0;

            vec3 skyColor = vec3(0.504, 0.700, 0.699);
            float skyColorMult = 1.0;
            skyColor *= skyColorMult;
            // 0.5 == sky gives some ambient light
            float skyDiffuseAmount = clamp(0.5 + 0.5 * dot(normal, vec3(0., 1., 0.)), 0.0, 1.0); // ??? why clamp?

            VAR color3 bounceColor = vec3(0.1, 0.1, 0.1);
            // 0.5 == bounce gives some ambient light
            float bounceDiffuseAmount = clamp(0.5 + 0.5 * dot(normal, vec3(0., -1., 0.)), 0.0, 1.0); // ??? why clamp?

            float sunDiffuseAmount = clamp(dot(normal, sunDirection), 0.0, 1.0); // ??? why clamp?
            vec3 sunHalfDir = normalize(sunDirection - ray);
            float sunSpecularAmount = pow(clamp(dot(reflect(ray, normal), sunDirection), 0.0, 1.0), materialSpecularity.y);

            // COMPILE TIME: disable shadow
#if 1
            HitInfo shadowHit = castRay(hitInfo.finalPos + 0.001 * normal, sunDirection);
            float sunShadow = shadowHit.rayHit == RAY_HIT ? 0.1 : 1.0;
#else
            float sunShadow = 1.0;
#endif

            //float sunSpecularAngle = max(dot(sunHalfDir, normal), 0.0);
            //float sunSpecularAmount = pow(sunSpecularAngle, shininess);

            vec3 light = skyDiffuseAmount * skyColor;
            light += bounceDiffuseAmount * bounceColor;
            light += sunDiffuseAmount * sunShadow * sunColor * (1. - materialSpecularity.x);
            light += sunSpecularAmount * sunShadow * sunSpecularColor * materialSpecularity.x;

            vec3 linearColor = light * material;
            color = pow(linearColor, vec3(1. / 2.2)); // gamma correct
#endif
        }
    }
    gl_FragColor = vec4(
           (smoothstep(0., 1., t)
         - smoothstep(9.9, 10., t)
         + smoothstep(10., 10.1, t)
         - smoothstep(19.9, 20., t)
         + smoothstep(20., 20.1, t)
         - smoothstep(30., 35., t)) * color, 1.0);
}
